@@ -4,7 +4,8 @@
 # ibus-anthy - The Anthy engine for IBus
 #
 # Copyright (c) 2007-2008 Peng Huang <shawn.p.huang@gmail.com>
-# Copyright (c) 2007-2011 Red Hat, Inc.
+# Copyright (c) 2010-2013 Takao Fujiwara <takao.fujiwara1@gmail.com>
+# Copyright (c) 2007-2013 Red Hat, Inc.
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -48,24 +49,50 @@ TYPING_MODE_KANA, \
 TYPING_MODE_THUMB_SHIFT = range(3)
 
 class JaString:
-    def __init__(self, mode=TYPING_MODE_ROMAJI):
-        self.__mode = mode
-        self.reset()
-        romaji.RomajiSegment._init_romaji_typing_rule(self._prefs)
-        kana.KanaSegment._init_kana_typing_rule(self._prefs)
-        thumb.ThumbShiftSegment._init_thumb_typing_rule(self._prefs)
+    _prefs = None
+    _mode = TYPING_MODE_ROMAJI
+    _shift = False
+    _unshift = False
 
-    def reset(self):
-        self.__cursor = 0
-        self.__segments = list()
-        self.__shift = False
+    def __init__(self, mode=TYPING_MODE_ROMAJI, latin_with_shift=True):
+        self._init_mode(mode)
+        if mode == TYPING_MODE_ROMAJI:
+            romaji.RomajiSegment.SET_LATIN_WITH_SHIFT(latin_with_shift)
 
-    def set_mode(self, mode):
-        self.__mode = mode
-        self.reset()
+    @classmethod
+    def _init_mode(cls, mode):
+        cls._mode = mode
+        cls._shift = False
+        cls._unshift = False
+        cls.__cursor = 0
+        cls.__segments = list()
+        if mode == TYPING_MODE_ROMAJI:
+            romaji.RomajiSegment._init_romaji_typing_rule(cls._prefs)
+        elif mode == TYPING_MODE_KANA:
+            kana.KanaSegment._init_kana_typing_rule(cls._prefs)
+        elif mode == TYPING_MODE_THUMB_SHIFT:
+            thumb.ThumbShiftSegment._init_thumb_typing_rule(cls._prefs)
+
+    @classmethod
+    def SET_PREFS(cls, prefs):
+        cls._prefs = prefs
+
+    @classmethod
+    def RESET(cls, prefs, section, name, value):
+        cls._prefs = prefs
+        if section.startswith('kana_typing_rule'):
+            mode = TYPING_MODE_KANA
+            kana.KanaSegment.RESET(prefs, section, name, value)
+            cls._init_mode(mode)
+        if section == 'common' and name == 'latin_with_shift':
+            romaji.RomajiSegment.SET_LATIN_WITH_SHIFT(value)
 
     def set_shift(self, shift):
-        self.__shift = shift
+        self._shift = shift
+
+    def set_hiragana_katakana(self, mode):
+        if mode and self._mode == TYPING_MODE_ROMAJI:
+            self._unshift = True
 
     def insert(self, c):
         segment_before = None
@@ -78,22 +105,32 @@ class JaString:
             segment_after = self.__segments[self.__cursor]
         if segment_before and not segment_before.is_finished():
             if type(segment_before) == romaji.RomajiSegment:
-                new_segments = segment_before.append(c, self.__shift)
+                new_segments = segment_before.append(c,
+                                                     self._shift,
+                                                     self._unshift)
+                self._unshift = False
             else:
                 new_segments = segment_before.append(c)
         elif segment_after and not segment_after.is_finished():
             if type(segment_after) == romaji.RomajiSegment:
-                new_segments = segment_after.prepend(c, self.__shift)
+                new_segments = segment_after.prepend(c,
+                                                     self._shift,
+                                                     self._unshift)
+                self._unshift = False
             else:
                 new_segments = segment_after.prepend(c)
         else:
-            if c != u"\0" and c != u"":
-                if self.__mode == TYPING_MODE_ROMAJI:
-                    new_segments = [romaji.RomajiSegment(c, u"", self.__shift)]
-                elif self.__mode == TYPING_MODE_KANA:
+            if c != u'\0' and c != u'':
+                if self._mode == TYPING_MODE_ROMAJI:
+                    new_segments = [romaji.RomajiSegment(c,
+                                                         u'',
+                                                         self._shift,
+                                                         self._unshift)]
+                    self._unshift = False
+                elif self._mode == TYPING_MODE_KANA:
                     # kana mode doesn't have shift latin in MS.
                     new_segments = [kana.KanaSegment(c)]
-                elif self.__mode == TYPING_MODE_THUMB_SHIFT:
+                elif self._mode == TYPING_MODE_THUMB_SHIFT:
                     new_segments = [thumb.ThumbShiftSegment(c)]
         if new_segments:
             self.__segments[self.__cursor:self.__cursor] = new_segments
@@ -212,7 +249,7 @@ class JaString:
         for c in s:
             c = c if not period else PeriodTable.get(c, c)
             # thumb_left + '2' and '/' are different
-            if self.__mode != TYPING_MODE_THUMB_SHIFT:
+            if self._mode != TYPING_MODE_THUMB_SHIFT:
                 c = c if not symbol else SymbolTable[symbol].get(c, c)
             c = c if not half_symbol else HalfSymbolTable.get(c, c)
             c = c if not half_number else HalfNumberTable.get(c, c)
@@ -222,34 +259,34 @@ class JaString:
     def get_hiragana(self, commit=False):
         conv = lambda s: s.to_hiragana()
         R = lambda s: s if not (commit and s[-1:] == u'n') else s[:-1] + u'ん'
-        text_before = R(u"".join(map(conv, self.__segments[:self.__cursor])))
-        text_after = R(u"".join(map(conv, self.__segments[self.__cursor:])))
+        text_before = R(u''.join(map(conv, self.__segments[:self.__cursor])))
+        text_after = R(u''.join(map(conv, self.__segments[self.__cursor:])))
         return self._chk_text(text_before + text_after), len(text_before)
 
     def get_katakana(self, commit=False):
         conv = lambda s: s.to_katakana()
         R = lambda s: s if not (commit and s[-1:] == u'n') else s[:-1] + u'ン'
-        text_before = R(u"".join(map(conv, self.__segments[:self.__cursor])))
-        text_after = R(u"".join(map(conv, self.__segments[self.__cursor:])))
+        text_before = R(u''.join(map(conv, self.__segments[:self.__cursor])))
+        text_after = R(u''.join(map(conv, self.__segments[self.__cursor:])))
         return self._chk_text(text_before + text_after), len(text_before)
 
     def get_half_width_katakana(self, commit=False):
         conv = lambda s: s.to_half_width_katakana()
         R = lambda s: s if not (commit and s[-1:] == u'n') else s[:-1] + u'ﾝ'
-        text_before = R(u"".join(map(conv, self.__segments[:self.__cursor])))
-        text_after = R(u"".join(map(conv, self.__segments[self.__cursor:])))
+        text_before = R(u''.join(map(conv, self.__segments[:self.__cursor])))
+        text_after = R(u''.join(map(conv, self.__segments[self.__cursor:])))
         return self._chk_text(text_before + text_after), len(text_before)
 
     def get_latin(self):
         conv = lambda s: s.to_latin()
-        text_before = u"".join(map(conv, self.__segments[:self.__cursor]))
-        text_after = u"".join(map(conv, self.__segments[self.__cursor:]))
+        text_before = u''.join(map(conv, self.__segments[:self.__cursor]))
+        text_after = u''.join(map(conv, self.__segments[self.__cursor:]))
         return text_before + text_after, len(text_before)
 
     def get_wide_latin(self):
         conv = lambda s: s.to_wide_latin()
-        text_before = u"".join(map(conv, self.__segments[:self.__cursor]))
-        text_after = u"".join(map(conv, self.__segments[self.__cursor:]))
+        text_before = u''.join(map(conv, self.__segments[:self.__cursor]))
+        text_after = u''.join(map(conv, self.__segments[self.__cursor:]))
         return text_before + text_after, len(text_before)
 
     def is_empty(self):
